/*------------------------------------------------------------------------------
	Newton interface to the native text decoding functions.
------------------------------------------------------------------------------*/

#define forDebug 1
#include <NewtonDebug.h>

#include <Objects.h>
#include <NewtonScript.h>
#include <Unicode.h>
#include <UStringUtils.h>

#ifdef __cplusplus
#define EXTERNC extern "C"
#else
#define EXTERNC
#endif


/*------------------------------------------------------------------------------
	C o n s t a n t s
------------------------------------------------------------------------------*/

#define kNoTranslationChar  0x1A
#define kEndOfUnicodeString 0x00


/*------------------------------------------------------------------------------
	T y p e s
------------------------------------------------------------------------------*/

// possible formats for the Unicode mapping tables
typedef UShort	TMapFormat;

// similar BUT NOT IDENTICAL to UnicodePrivate.h definitions
// these in fact match the resources
typedef struct
{
	TMapFormat	fMapFormatID;
	UShort		fTableInfo;		// sould be ULong, but is zero anyway
	ULong			fTableSize;		// should be UShort but compiler won't pack this struct properly
	UniChar		fCharCodes[1];	//	array size 0 not allowed by ARMcpp
} TEncodingMap;

typedef struct
{
	TMapFormat	fMapFormatID;
	UShort		fTableSize;		// could be a packing issue here, but this is ignored anyway
	ULong			fTableInfo;
	UniChar		fCharCodes[1];	//	array size 0 not allowed by ARMcpp
} TDecodingMap;


/*------------------------------------------------------------------------------
	F u n c t i o n   P r o t o t y p e s
------------------------------------------------------------------------------*/

static void ConvertToUnicode(const void * str, UniChar * ustr, void * table, FastInt numChars);
static void ConvertFromUnicode(const UniChar * ustr, void * str, void * table, FastInt numChars);


/*------------------------------------------------------------------------------
	Convert Unicode to an ISO-8859-1 encoded string.

	In:		theText
				Unicode string
				theData
				binary object large enough to hold converted ISO-8859-1 encoded data
				theTable
				binary object describing encoding

	Return:	binary data
------------------------------------------------------------------------------*/

EXTERNC Ref Encode(RefArg rcvr, RefArg theText, RefArg theData, RefArg theTable)
{
	//	theText must be a string
	if (!IsString(theText))
		ThrowBadTypeWithFrameData(kNSErrNotAString, theText);
	//	theData must be binary
	if (!IsBinary(theData))
		ThrowBadTypeWithFrameData(kNSErrNotABinaryObject, theData);
	//	theTable must be binary
	if (!IsBinary(theTable))
		ThrowBadTypeWithFrameData(kNSErrNotABinaryObject, theTable);

	WITH_LOCKED_BINARY(theText, textPtr)
	{
		WITH_LOCKED_BINARY(theData, dataPtr)
		{
			WITH_LOCKED_BINARY(theTable, tablePtr)
			{

				FastInt numChars = Length(theText) / 2 - 1;
				ConvertFromUnicode((UniChar *)textPtr, dataPtr, tablePtr, numChars);

			} END_WITH_LOCKED_BINARY(theTable);
		} END_WITH_LOCKED_BINARY(theData);
	} END_WITH_LOCKED_BINARY(theText);

	return theData;
}


/*------------------------------------------------------------------------------
	Convert ISO-8859-1 (Byte) data to Unicode string.

	In:		theData
				ISO-8859-1 encoded binary data, NOT NUL terminated
				theText
				Unicode string large enough to hold converted data
				theTable
				binary object describing decoding

	Return:	Unicode string
------------------------------------------------------------------------------*/

EXTERNC Ref Decode(RefArg rcvr, RefArg theData, RefArg theText, RefArg theTable)
{
	//	theData must be binary
	if (!IsBinary(theData))
		ThrowBadTypeWithFrameData(kNSErrNotABinaryObject, theData);
	//	theText must be a string
	if (!IsString(theText))
		ThrowBadTypeWithFrameData(kNSErrNotAString, theText);
	//	theTable must be binary
	if (!IsBinary(theTable))
		ThrowBadTypeWithFrameData(kNSErrNotABinaryObject, theTable);

	WITH_LOCKED_BINARY(theData, dataPtr)
	{
		WITH_LOCKED_BINARY(theText, textPtr)
		{
			WITH_LOCKED_BINARY(theTable, tablePtr)
			{
				FastInt numChars = Length(theData);
				ConvertToUnicode(dataPtr, (UniChar *)textPtr, tablePtr, numChars);
				*((UniChar *)textPtr + numChars) = kEndOfUnicodeString;

			} END_WITH_LOCKED_BINARY(theTable);
		} END_WITH_LOCKED_BINARY(theText);
	} END_WITH_LOCKED_BINARY(theData);

	return theText;
}


/*------------------------------------------------------------------------------
	Generic Byte data to Unicode string conversion function.

	In:		str
				character string
				ustr
				Unicode string large enough to hold converted data
				table
				encoding -> Unicode mapping table
				numChars
				number of characters to convert

	Return:	void
------------------------------------------------------------------------------*/

static void ConvertToUnicode(const void * str, UniChar * ustr, void * table, FastInt numChars)
{
	const unsigned char *	p = (const unsigned char *)str;

	while (numChars-- > 0)
	{
		*ustr++ = ((TDecodingMap *)table)->fCharCodes[*p++];
	}
}


/*------------------------------------------------------------------------------
	Generic Unicode string to Byte data conversion function.

	In:		ustr
				Unicode string
				str
				character string large enough to hold converted data
				table
				Unicode -> encoding mapping table
				numChars
				number of characters to convert

	Return:	void
------------------------------------------------------------------------------*/

static void ConvertFromUnicode(const UniChar * ustr, void * str, void * table, FastInt numChars)
{
	FastInt		tableSize  = ((TEncodingMap *)table)->fTableSize;
	UniChar *	tablePtr   = ((TEncodingMap *)table)->fCharCodes;
	UniChar *	endCodes   = tablePtr;											/* end Unicode of run */
	UniChar *	startCodes = tablePtr + tableSize;							/* start Unicode of run */
	SShort *		codeDeltas = (SShort *)(tablePtr + tableSize * 2);		/* offset from Unicode into codeArray */
	char *		charCodes  = (char *)(tablePtr + tableSize * 3);		/* 8-bit char corresponding to Unicode */
	char *		p = (char *)str;

	while (numChars-- > 0)
	{
		UniChar ch = kNoTranslationChar;
		UniChar uch = *ustr++;
		for (FastInt i = 0; i < tableSize; i++)
		{
			if (uch >= startCodes[i]
			 && uch <= endCodes[i])
			{
				ch = charCodes[uch + codeDeltas[i]];
				break;
			}
		}
		*p++ = ch;
	}
}
