/*------------------------------------------------------------------------------
	Newton interface to the native text decoding functions.
------------------------------------------------------------------------------*/

//#define forDebug 1
//#include "NewtonDebug.h"

#include "Objects.h"
#include "NewtonScript.h"
#include "Unicode.h"
#include "UStringUtils.h"

#ifdef __cplusplus
#define EXTERNC extern "C"
#else
#define EXTERNC
#endif

#if defined(__MWERKS__)
#undef  WITH_LOCKED_BINARY
#define WITH_LOCKED_BINARY(r,p) void*p=(void*)r;
#undef  END_WITH_LOCKED_BINARY
#define END_WITH_LOCKED_BINARY(r)
#endif


/*------------------------------------------------------------------------------
	C o n s t a n t s
------------------------------------------------------------------------------*/

#define kEndOfUnicodeString 0x00


/*------------------------------------------------------------------------------
	Strip leading dots from lines in text.

	In:		theText
				string ref containing RFC-822 message text which may have escaped
				dots at the start of lines

	Return:	modified string
------------------------------------------------------------------------------*/

EXTERNC Ref Dots(RefArg rcvr, RefArg theText)
{
	//	parameter must be a string
	if (!IsString(theText))
		ThrowBadTypeWithFrameData(kNSErrNotAString, theText);

	WITH_LOCKED_BINARY(theText, textPtr)
	{
		UniChar * srcPtr = (UniChar *)textPtr;
		UniChar * destPtr = srcPtr;
		UniChar * stopPtr = srcPtr + (Length(theText) / 2);
		bool newline = true;
		for ( ; srcPtr < stopPtr; srcPtr++)
		{
			if (newline && (*srcPtr == '.'))
				;	//	do nothing
			else
				*destPtr++ = *srcPtr;
			newline = (*srcPtr == '\n');
		}
		if (srcPtr != destPtr)
		{
			NSCallGlobalFn(SYM(SetLength), theText, MakeInt((long)destPtr - (long)textPtr));
		}
	} END_WITH_LOCKED_BINARY(theText);

	return theText;
}


/*------------------------------------------------------------------------------
	The unicode -> ISO-8859-1 mapping tables.
------------------------------------------------------------------------------*/

const ULong kNumOfRuns = 25;
const struct ISOmap {
	ULong numOfRuns;
	UniChar endCodeArray[kNumOfRuns];		/* end Unicode of run */
	UniChar startCodeArray[kNumOfRuns];		/* start Unicode of run */
	SShort codeDeltaArray[kNumOfRuns];		/* offset from Unicode into codeArray */
	UChar codeArray[294];						/* 8-bit char corresponding to Unicode */
}
mapUnicodeToISO = {
	kNumOfRuns,
#include "Map To ISO-8859-1"
};


/*------------------------------------------------------------------------------
	The ISO-8859-1 -> unicode mapping table.
------------------------------------------------------------------------------*/

const UniChar mapISOtoUnicode[256] =
#include "Map From ISO-8859-1"
;


/*------------------------------------------------------------------------------
	Convert Unicode to ISO-8859-1.

	In:		theChar
				integer ref value of Unicode character

	Return:	integer ref : ISO-8859-1 equivalent
------------------------------------------------------------------------------*/

EXTERNC Ref EncodeISOchar(RefArg rcvr, RefArg theChar)
{
	//	parameter must actually be an integer
	if (!IsInt(theChar))
		ThrowBadTypeWithFrameData(kNSErrNotAnInteger, theChar);

	UniChar result = kNoTranslationChar;
	UniChar ch = RefToInt(theChar);
	for (FastInt i = 0; i < mapUnicodeToISO.numOfRuns; i++)
	{
		if ((ch >= mapUnicodeToISO.startCodeArray[i])
		 && (ch <= mapUnicodeToISO.endCodeArray[i]))
		{
			ch += mapUnicodeToISO.codeDeltaArray[i];
			result = mapUnicodeToISO.codeArray[ch];
			break;
		}
	}

	return MakeInt(result);
}


/*------------------------------------------------------------------------------
	Convert ISO-8859-1 (UniChar) string to unicode.

	In:		theText
				string ref containing ISO-8859-1 encoded characters

	Return:	modified string
------------------------------------------------------------------------------*/

EXTERNC Ref DecodeISO88591(RefArg rcvr, RefArg theText)
{
	//	parameter must be a string
	if (!IsString(theText))
		ThrowBadTypeWithFrameData(kNSErrNotAString, theText);

	WITH_LOCKED_BINARY(theText, textPtr)
	{
		UniChar * srcPtr = (UniChar *)textPtr;
		UniChar * stopPtr = srcPtr + (Length(theText) / 2);
		for ( ; srcPtr < stopPtr; srcPtr++)
		{
			*srcPtr = mapISOtoUnicode[*srcPtr & 0x00FF];
		}
	} END_WITH_LOCKED_BINARY(theText);

	return theText;
}



/*------------------------------------------------------------------------------
	Convert ISO-8859-1 (Byte) data to unicode string.

	In:		theData
				ISO-8859-1 binary data
				theText
				string large enough to hold converted data

	Return:	unicode string
------------------------------------------------------------------------------*/

EXTERNC Ref ToISO88591(RefArg rcvr, RefArg theData, RefArg theText)
{
	//	theData must be binary
	if (!IsBinary(theData))
		ThrowBadTypeWithFrameData(kNSErrNotABinaryObject, theData);
	//	theText must be a string
	if (!IsString(theText))
		ThrowBadTypeWithFrameData(kNSErrNotAString, theText);

	WITH_LOCKED_BINARY(theData, dataPtr)
	{
		WITH_LOCKED_BINARY(theText, textPtr)
		{
			UByte * srcPtr = (UByte *)dataPtr;
			UniChar * destPtr = (UniChar *)textPtr;
			FastInt numChars = Length(theData) - 1;
			while (numChars-- > 0)
			{
				*destPtr++ = mapISOtoUnicode[*srcPtr++];
			}
			*destPtr = 0;
		} END_WITH_LOCKED_BINARY(theText);
	} END_WITH_LOCKED_BINARY(theData);

	return theText;
}



/*------------------------------------------------------------------------------
	Convert unicode string to byte data. Just strip the top byte of each unicode
	char.

	In:		theText
				string
				theData
				binary object large enough to hold the data

	Return:	the data
------------------------------------------------------------------------------*/

EXTERNC Ref ToData(RefArg rcvr, RefArg theText, RefArg theData)
{
	//	theText must be a string
	if (!IsString(theText))
		ThrowBadTypeWithFrameData(kNSErrNotAString, theText);
	//	theData must be binary
	if (!IsBinary(theData))
		ThrowBadTypeWithFrameData(kNSErrNotABinaryObject, theData);

	WITH_LOCKED_BINARY(theText, textPtr)
	{
		WITH_LOCKED_BINARY(theData, dataPtr)
		{
			UniChar * srcPtr = (UniChar *)textPtr;
			char * destPtr = (char *)dataPtr;
			FastInt numChars = Length(theText) / 2 - 1;
			while (numChars-- > 0)
			{
				*destPtr++ = (char) *srcPtr++;
			}
		} END_WITH_LOCKED_BINARY(theData);
	} END_WITH_LOCKED_BINARY(theText);

	return theData;
}



/*------------------------------------------------------------------------------
	Convert byte data to unicode string. Just copy with no unicode conversion.

	In:		theData
				an array object containing byte data
				theText
				string large enough to hold the data
				theCount
				number of characters to copy

	Return:	the text
------------------------------------------------------------------------------*/

EXTERNC Ref ToString(RefArg rcvr, RefArg theData, RefArg theText, RefArg theCount)
{
	//	theData must be an array of bytes
	if (!IsArray(theData))
		ThrowBadTypeWithFrameData(kNSErrNotAnArray, theData);
	//	theText must be a string
	if (!IsString(theText))
		ThrowBadTypeWithFrameData(kNSErrNotAString, theText);
	//	theCount must be Dracula himself... er, no, an integer
	if (!IsInt(theCount))
		ThrowBadTypeWithFrameData(kNSErrNotAnInteger, theCount);

	FastInt numChars = RefToInt(theCount);
	if (numChars > 0)
	{
		WITH_LOCKED_BINARY(theData, dataPtr)
		{
			WITH_LOCKED_BINARY(theText, textPtr)
			{
				Ref * srcPtr = (Ref *)dataPtr;
				UniChar * destPtr = (UniChar *)textPtr;
				while (numChars-- > 0)
				{
					*destPtr++ = (UniChar)RefToInt(*srcPtr++);
				}
				*destPtr = kEndOfUnicodeString;
			} END_WITH_LOCKED_BINARY(theText);
		} END_WITH_LOCKED_BINARY(theData);
	}

	return theText;
}



/*------------------------------------------------------------------------------
	Convert quoted-printable string to UniChar.

	In:		theText
				string containing quoted-printable escaped characters
				doMapISO
				map to ISO-8859-1 after unescaping characters flag

	Return:	modified string
------------------------------------------------------------------------------*/


static inline short ToHex(UniChar ch)
{
	short x = ch - 0x30;		//	base on 0
	if (x > 9)
		x -= 7;					//	handle A-F
	return x;
}


EXTERNC Ref DecodeQuotedPrintable(RefArg rcvr, RefArg theText, RefArg doMapISO)
{
	//	parameter must be a string
	if (!IsString(theText))
		ThrowBadTypeWithFrameData(kNSErrNotAString, theText);

	bool doMap = NOTNIL(doMapISO);

	WITH_LOCKED_BINARY(theText, textPtr)
	{
		UniChar * srcPtr = (UniChar *)textPtr;
		UniChar * stopPtr = srcPtr + (Length(theText) >> 1);
		for ( ; srcPtr < stopPtr; srcPtr++)
		{
			if (*srcPtr == '=')
				break;
		}

		enum
		{
			kNothing,
			kQuote,
			kQuoted1
		} quotedState = kNothing;
		short quotedChar = 0;

		UniChar *destPtr = srcPtr;
		for ( ; srcPtr < stopPtr; srcPtr++)
		{
			switch (quotedState)
			{
			case kNothing:
				if  (*srcPtr == '=')
				{
					quotedState = kQuote;
				}
				else if  (*srcPtr == '_')
				{
					*destPtr++ = ' ';
				}
				else
				{
					*destPtr++ = *srcPtr;
				}
				break;

			case kQuote:
				if (*srcPtr == '\n')
				{
					quotedState = kNothing;			//	unfold "=\n"
				}
				else
				{
					quotedChar = ToHex(*srcPtr);
					quotedState = kQuoted1;
				}
				break;

			case kQuoted1:
				quotedChar = (quotedChar << 4) + ToHex(*srcPtr);
				*destPtr++ = doMap ? mapISOtoUnicode[quotedChar] : quotedChar;
				quotedState = kNothing;
			}
		}
		if (srcPtr != destPtr)
		{
			NSCallGlobalFn(SYM(SetLength), theText, MakeInt((long)destPtr - (long)textPtr));
		}
	} END_WITH_LOCKED_BINARY(theText);

	return theText;
}
