/*------------------------------------------------------------------------------
	Function to restore object addresses to offsets within a package.
------------------------------------------------------------------------------*/

//#define forDebug 1
//#include "NewtonDebug.h"

#include "Objects.h"
#include "PackageSpec.h"

#ifdef __cplusplus
#define EXTERNC extern "C"
#else
#define EXTERNC
#endif


/*----------------------------------------------------------------------
	R e f   T a g   B i t s
----------------------------------------------------------------------*/

enum
{
	kIntegerTag,
	kPointerTag,
	kImmediateTag,
	kMagicPointerTag,

	kTagMask = 0x03
};

typedef enum
{
	kImmedSpecial,
	kImmedChar,
	kImmedBoolean,
	kImmedReserved,

	kImmedMask = 0x03
} ImmediateType;


/*----------------------------------------------------------------------
	O b j e c t   T a g   F u n c t i o n s
----------------------------------------------------------------------*/

inline long Tag(Ref ref) { return ref & kTagMask; }

#define ISINT(_r) (Tag(_r) == kIntegerTag)
#define NOTINT(_r) (Tag(_r) != kIntegerTag)

#define INTR(_i) ((_i) << 2)
#define RINT(_r) ((_r) >> 2)

#define RPTR(_o) ((char*)((Ref)_o & ~0x03))
#define PTRR(_o) ((Ref)((char*)_o + 1))


/*----------------------------------------------------------------------
	N e w t o n   O b j e c t s
----------------------------------------------------------------------*/

class ObjHeader
{
public:
	unsigned size  : 24;
	unsigned flags :  8;

	unsigned	lockCount :  8;
	unsigned	slotIndex : 24;

};

// ARMcpp doesnt allow zero array sizes so well have to define them as 1
// and allow for this where appropriate

class ArrayObject : public ObjHeader
{
public:
	Ref objClass;
	Ref slot[1];
};

//	Newton object flags
typedef enum
{
	kObjSlotted  = 0x01,
	kObjFrame    = 0x02,
	kObjFree     = 0x04,
	kObjMarked   = 0x08,
	kObjLocked   = 0x10,
	kObjForward  = 0x20,
	kObjReadOnly = 0x40,
	kObjDirty    = 0x80,

	kBinaryObject = 0x00,
	kLargeBinaryObject = kObjFrame,
	kArrayObject = kObjSlotted,
	kFrameObject = kObjSlotted + kObjFrame,
	kObjMask     = kObjSlotted + kObjFrame
} ObjFlags;


/*------------------------------------------------------------------------------
	Restore object addresses to offsets within a package.
	We do this by following object refs from each part base and fixing up
	the addresses.
	Input:
		package
	Output:
		modified package
------------------------------------------------------------------------------*/


//	Determine whether a Newton ref is a pointer (could be an immediate)

inline Boolean IsPtr(Ref ref)
{
	return (Tag(ref) == kPointerTag) && ((ref & 0xFF000000) != 0x00);
}


//	Fix up the pointer ref offset from base by delta

static void FixRef(Ptr base, Ref ref, ULong delta)
{
//	point to the object - could  be frame or array as well as binary
	ArrayObject * rObj = (ArrayObject *)(base + (ULong) RPTR(ref));
	Ref fixedRef;

//	point to its class
	Ref * refPointer = &(rObj->objClass);
	if (IsPtr(*refPointer))
	{
		fixedRef = (Ref) (RPTR(*refPointer) - delta);
		*refPointer = PTRR(fixedRef);						//	update the class ref
		if (rObj->flags == (kObjReadOnly + kFrameObject))
			FixRef(base, fixedRef, delta);				//	for frames, class is actually map
	}

//	if its a frame / array, step through each slot / element fixing those
	if ((rObj->flags & (0x00FF & ~kObjFrame)) == (kObjReadOnly + kObjSlotted))
	{
		refPointer = &rObj->slot[0];
		for (FastInt refCounter = (rObj->size - sizeof(ArrayObject)) / sizeof(Ref);  refCounter >= 0; refCounter--, refPointer++)
		{
			if (IsPtr(*refPointer))
			{
				fixedRef = (Ref) (RPTR(*refPointer) - delta);
				if (fixedRef > 0)
				{
					*refPointer = PTRR(fixedRef);		//	update this ref
					FixRef(base, fixedRef, delta);	//	follow it to update sub-refs
				}
			}
		}
	}
}


#if defined(__MWERKS__)
#undef  WITH_LOCKED_BINARY
#define WITH_LOCKED_BINARY(r,p) void*p=(void*)r;
#undef  END_WITH_LOCKED_BINARY
#define END_WITH_LOCKED_BINARY(r)
#endif


EXTERNC Ref UnRelocatePkg(RefArg rcvr, RefArg pkg)
{
	//	parameter must be a binary object
	if (!IsBinary(pkg))
		ThrowBadTypeWithFrameData(kNSErrNotAPackage, pkg);

	WITH_LOCKED_BINARY(pkg, p)
	{
		Ptr pp = (Ptr)p;
		PackageDirectory * dir = (PackageDirectory *)p;
		ULong partData = dir->directorySize;
	//	if its a "package1" with relocation info then skip that relocation info
		if ((dir->signature[7] == '1') && (dir->flags & kRelocationFlag))
			partData += ((RelocationHeader *)(pp + partData))->relocationSize;

	//	calculate delta to be applied to refs
		ArrayObject * rObj = (ArrayObject *)(pp + partData);
		ULong delta = (ULong)RPTR(rObj->slot[0])				//	address of first object
							- (partData + rObj->size);				//	offset of first object

	//	fix up refs for all parts
		for (FastInt partNumber = 0; partNumber < dir->numParts; partNumber++)
			FixRef(pp, partData + dir->parts[partNumber].offset, delta);

	} END_WITH_LOCKED_BINARY(pkg);

	return pkg;
}
